////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "cycles.h"
#include "crnlp_uniform_fgls.h"
#include "nesting.h"
#include "mem.h"
#include "queue.h"

void crnlp_uniform_fgls_init(crnlp_uniform_fgls_lock_t* lock)
{
    int i = 0;

    lock->head = 0;

    for (i=0; i < SCHEDSIZE; i++)    { lock->enabled[i] = 0; }
    for (i=0; i < SCHEDSIZE; i++)    { lock->blocked[i] = 0; }
    for (i=0; i < SCHEDSIZE; i++)    { lock->schedule[i] = 0; }
    for (i=0; i < QUEUE_LENGTH; i++) { lock->requests[i].done = 0;}
    for (i=0; i < QUEUE_LENGTH; i++) { lock->requests[i].needs_service = 0;}
    for (i=0; i < QUEUE_LENGTH; i++) { lock->requests[i].spin_location = -1;} // invalid

    lock->enabled[lock->head] = 1;

    lock->pending_requests = 0;

    lock->service_index = 0;
    lock->server_exists = 0;
}

uint64_t *crnlp_uniform_fgls_lock(crnlp_uniform_fgls_lock_t* lock, uint8_t ind)
{
    crnlp_uniform_fgls_node_t *curr_node, *self, *other;
    uint8_t start_in_sched, i;
    uint8_t is_lock_server;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    self = &(lock->requests[ind]);

    // Mark this request as needing to be serviced for a lock
    self->done = 0;
    mb();
    self->needs_service = 1;

    // Check whether a lock server exists (and potentially become it)
    is_lock_server = 0;
    if (!lock->server_exists)
    {
        is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
    }

    // Wait until this request is serviced, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will service itself later)
    while (!is_lock_server && self->needs_service != 0)
    {
        // Try to become the lock server (in the case the previous lock server
        // completed without servicing this request)
        if (!lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
        }
    }

    // If this core is not the lock server, then we switch to measuring
    // blocking time (the request must have been serviced)
    if (!is_lock_server)
    {
        end = get_cycles();
        ret[0] += end - start;
        start = get_cycles();
    }

    // Wait until this request is satisfied, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will eventually become
    // satisfied, after potentially servicing other requests)
    while (!is_lock_server && lock->enabled[self->spin_location] != 1) 
    {
        // Try to become the lock server (in order to help service requests
        // while waiting to be satisfied)
        if (!lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
        }
    }

    // If this core is not the lock server, then the request must be satisfied
    if (!is_lock_server)
    {
        end = get_cycles();
        ret[1] += end - start;

        return ret;
    }

    // Otherwise, help out as long as this request has not yet been serviced
    // and is not yet satisfied
    while (self->needs_service != 0 || lock->enabled[self->spin_location] != 1)
    {
        // Check if the next request needs to be serviced
        i = lock->service_index;
        if (lock->requests[i].needs_service == 1)
        {
            // Service request i
            other = &(lock->requests[i]);
            if (other->done == 0)
            {
                // Lock code -- determine which row in the table in which
                // to place this request
                if (lock->pending_requests > 0)
                {
                    start_in_sched = (lock->head + 1) % SCHEDSIZE;
                    while ((lock->schedule[start_in_sched] & other->requested) != 0)
                    {
                        start_in_sched = (start_in_sched + 1) % SCHEDSIZE;
                    }
                }
                else
                {
                    start_in_sched = lock->head;
                }

                // Set the next value for the request
                other->next = (start_in_sched + 1) % SCHEDSIZE;

                // Update blocked for the next row, the count of pending
                // requests, and set the bits in this row corresponding to the
                // resources needed by the request
                lock->blocked[other->next] += 1;
                lock->pending_requests += 1;
                lock->schedule[start_in_sched] |= other->requested;

                // Set the spin location for the request to be its row in the
                // table; it will be satisfied when enabled is set to true
                // for this row
                other->spin_location = start_in_sched;

                mb();

                // Mark the request as serviced
                other->needs_service = 0;
            }
            else
            {
                // Unlock code - clear the bits in the table corresponding to
                // the resources that were needed for the request
                lock->schedule[lock->head] &= ~(other->requested);

                // Decrement blocked and the pending request count, as the
                // request is now finished
                lock->blocked[other->next] -= 1;
                lock->pending_requests -= 1;

                // If this was the last request blocking the next row, then
                // mark the next row as enabled
                if (lock->blocked[other->next] == 0)
                {
                    lock->enabled[lock->head] = 0;
                    lock->head = other->next;
                    lock->enabled[lock->head] = 1;
                }

                mb();

                // Mark the request as serviced
                other->needs_service = 0;
            }

            // If this request was just serviced, then we switch to measuring
            // blocking time
            if (i == ind)
            {
                end = get_cycles();
                ret[0] += end - start;
                start = get_cycles();
            }
        }

        // Move onto the next request
        lock->service_index = (i + 1) % QUEUE_LENGTH;
    }

    // Set the server to not exist
    lock->server_exists = 0;

    end = get_cycles();
    ret[1] += end - start;
    start = get_cycles();

    // Satisfied!
    return ret;
}

uint64_t *crnlp_uniform_fgls_unlock(crnlp_uniform_fgls_lock_t* lock, uint8_t ind)
{
    crnlp_uniform_fgls_node_t *curr_node, *self;
    uint8_t is_lock_server;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    self = &(lock->requests[ind]);

    // Mark this request as needing to be serviced for an unlock
    self->done = 1;
    mb();
    self->needs_service = 1;

    // Check whether a lock server exists (and potentially become it)
    is_lock_server = 0;
    if (!lock->server_exists)
    {
        is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
    }

    // Wait until this request is serviced, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will service itself later)
    while (!is_lock_server && self->needs_service != 0)
    {
        // Try to become the lock server (in the case the previous lock server
        // completed without servicing this request)
        if (!lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
        }
    }

    // If this core is the lock server and has not been serviced, be
    // greedy and just service this unlock
    if (is_lock_server && self->needs_service == 0)
    {
        // Oops, this core shouldn't be the server
        lock->server_exists = 0;
    }
    else if (is_lock_server)
    {
        // Clear the bits in the table
        // corresponding to the resources that were needed for this request
        lock->schedule[lock->head] &= ~(self->requested);

        // Decrement blocked and the pending request count, as this
        // request is now finished
        lock->blocked[self->next] -= 1;
        lock->pending_requests -= 1;

        // If this was the last request blocking the next row, then
        // mark the next row as enabled
        if (lock->blocked[self->next] == 0)
        {
            lock->enabled[lock->head] = 0;
            lock->head = self->next;
            lock->enabled[lock->head] = 1;
        }

        mb();

        // Mark this request as serviced
        self->needs_service = 0;

        // Set the server to not exist
        lock->server_exists = 0;
    }

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

