#include <stdio.h>
#include <stdlib.h>

#include "cycles.h"
#include "r7lp.h"

///////////////////////////////////////////////////////////////////////
// Definitions of BITS, PHID, and PRES for different groups          //
///////////////////////////////////////////////////////////////////////

const unsigned int r7lp_bits[] = { 0x00000003, // 0000 0000 0000 0000 0000 0000 0000 0011
                                   0x0000000c, // 0000 0000 0000 0000 0000 0000 0000 1100
                                   0x00000030,
                                   0x000000c0,
                                   0x00000300,
                                   0x00000c00,
                                   0x00003000,
                                   0x0000c000,
                                   0x00030000,
                                   0x000c0000,
                                   0x00300000,
                                   0x00c00000,
                                   0x03000000,
                                   0x0c000000,
                                   0x30000000,
                                   0xc0000000 };

const unsigned int r7lp_phid[] = { 0x00000001, // 0000 0000 0000 0000 0000 0000 0000 0001
                                   0x00000004, // 0000 0000 0000 0000 0000 0000 0000 0100
                                   0x00000010,
                                   0x00000040,
                                   0x00000100,
                                   0x00000400,
                                   0x00001000,
                                   0x00004000,
                                   0x00010000,
                                   0x00040000,
                                   0x00100000,
                                   0x00400000,
                                   0x01000000,
                                   0x04000000,
                                   0x10000000,
                                   0x40000000 };

const unsigned int r7lp_pres[] = { 0x00000002, // 0000 0000 0000 0000 0000 0000 0000 0010
                                   0x00000008, // 0000 0000 0000 0000 0000 0000 0000 1000
                                   0x00000020,
                                   0x00000080,
                                   0x00000200,
                                   0x00000800,
                                   0x00002000,
                                   0x00008000,
                                   0x00020000,
                                   0x00080000,
                                   0x00200000,
                                   0x00800000,
                                   0x02000000,
                                   0x08000000,
                                   0x20000000,
                                   0x80000000 };

///////////////////////////////////////////////////////////////////////
// Initialization                                                    //
///////////////////////////////////////////////////////////////////////

void r7lp_init(r7lp_t *lock)
{
    unsigned int i;

    for (i = 0; i < R7LP_K; i++)
    {
        lock->groups[i].in = lock->groups[i].out = 0;
	lock->groups[i].sat = lock->groups[i].head = 0;
	lock->groups[i].wave = 0;

        lock->cached_state[i].S = 0;
        lock->cached_state[i].is_valid = 0;
    }

    lock->S = 0;
}

///////////////////////////////////////////////////////////////////////
// Lock                                                              //
///////////////////////////////////////////////////////////////////////

uint64_t *r7lp_lock(r7lp_t *lock, unsigned int groupnum)
{
    unsigned int a, b, c, d, e, f, g;
    unsigned int bcdefg;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0; // [overhead, blocking]

    start = get_cycles();

    // Take a ticket
    a = __sync_fetch_and_add(&lock->groups[groupnum].in, 1);

    end = get_cycles();
    ret[0] += end - start;
    start = get_cycles();

    // Wait until head or satisfied
    while (lock->groups[groupnum].head != a)
    {
        if (lock->groups[groupnum].sat >= a)
        {
            // We got the lock!  (Also, we're not responsible for cleanup.)
            end = get_cycles();
            ret[1] += end - start;

            return ret;
        }

        cpu_relax();
    }

    end = get_cycles();
    ret[1] += end - start;
    start = get_cycles();

    // If we're here, we need to do some work - check if the lock state
    // is cached and if so, read it; if not, update the phase and present
    // bit and read in the lock state
    if (lock->cached_state[groupnum].is_valid == 1)
    {
        bcdefg = lock->cached_state[groupnum].S;

        lock->cached_state[groupnum].is_valid = 0;
    }
    else
    {
        lock->groups[groupnum].wave = ~lock->groups[groupnum].wave;

        bcdefg = __sync_fetch_and_add(&lock->S, r7lp_pres[groupnum] | (lock->groups[groupnum].wave & r7lp_phid[groupnum]));
    }

    b = bcdefg & r7lp_bits[(groupnum + 1) % R7LP_K];
    c = bcdefg & r7lp_bits[(groupnum + 2) % R7LP_K];
    d = bcdefg & r7lp_bits[(groupnum + 3) % R7LP_K];
    e = bcdefg & r7lp_bits[(groupnum + 4) % R7LP_K];
    f = bcdefg & r7lp_bits[(groupnum + 5) % R7LP_K];
    g = bcdefg & r7lp_bits[(groupnum + 6) % R7LP_K];

    end = get_cycles();
    ret[0] += end - start;
    start = get_cycles();

    // Wait until prior phases are complete
    while (((b != 0) && (b == (lock->S & (r7lp_bits[(groupnum + 1) % R7LP_K])))) || 
           ((c != 0) && (c == (lock->S & (r7lp_bits[(groupnum + 2) % R7LP_K])))) ||
           ((d != 0) && (d == (lock->S & (r7lp_bits[(groupnum + 3) % R7LP_K])))) ||
           ((e != 0) && (e == (lock->S & (r7lp_bits[(groupnum + 4) % R7LP_K])))) ||
           ((f != 0) && (f == (lock->S & (r7lp_bits[(groupnum + 5) % R7LP_K])))) ||
           ((g != 0) && (g == (lock->S & (r7lp_bits[(groupnum + 6) % R7LP_K])))))
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[1] += end - start;
    start = get_cycles();

    // Update sat so all other requests in this phase know they're satisfied
    lock->groups[groupnum].sat = lock->groups[groupnum].in - 1;

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

///////////////////////////////////////////////////////////////////////
// Unlock                                                            //
///////////////////////////////////////////////////////////////////////

uint64_t *r7lp_unlock(r7lp_t *lock, unsigned int groupnum)
{
    unsigned int a;
    unsigned int cached_s;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0; // [overhead, blocking]

    start = get_cycles();

    // Update out to see if we're the last of the phase
    a = __sync_fetch_and_add(&lock->groups[groupnum].out, 1);

    // If this request is the last in the phase, modify S to mark this
    // phase complete, and update the head ticket
    if (a == lock->groups[groupnum].sat)
    {
        // To avoid a race condition when another request in this
        // group is already present (but must wait until the next
        // time this group is satisfied), do not clear this group's
        // bits in S.  Rather, update the phase and store the
        // current S information for the next request to pick up.
        if (lock->groups[groupnum].in - 1 > lock->groups[groupnum].sat)
        {
            lock->groups[groupnum].wave = ~lock->groups[groupnum].wave;

            // Switch the phase bit in S for this group, but leave
            // the present bit set
            if (lock->groups[groupnum].wave != 0)
            {
                cached_s = __sync_fetch_and_or(&lock->S, r7lp_phid[groupnum]);
            }
            else
            {
                cached_s = __sync_fetch_and_and(&lock->S, ~r7lp_phid[groupnum]);
            }

            lock->cached_state[groupnum].S = cached_s;
            lock->cached_state[groupnum].is_valid = 1;
        }
        else
        {
            __sync_fetch_and_and(&lock->S, ~r7lp_bits[groupnum]);
        }

        lock->groups[groupnum].head = a + 1;
    }

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

