#include "sharedres.h"
#include "blocking.h"
#include "rw-blocking.h"
#include <iostream>

#include "stl-helper.h"

BlockingBounds* cglp_bounds(const ResourceSharingInfo& info,
                            unsigned int procs_per_cluster,
                            int dedicated_irq)
{
    // Split everything by partition
    Clusters clusters;
    split_by_cluster(info, clusters);

    // Split each partition by resource
    ClusterResources resources; // std::vector<Resources>
    split_by_resource(clusters, resources);

    // Sort each contention set by request length
    // (resources will have one cluster and thus one element,
    //  which is effectively an array of groups, and each
    //  group knows the requests belonging to it)
    sort_by_request_length(resources);

    // Filter out the longest request for each group into group_reqs_first
    std::vector<const RequestBound*> group_reqs_first;
    unsigned int num_groups = 0;
    unsigned int k = 0;
    unsigned int max_resource_id = 0;
    for(k=0; k<resources[0].size(); k++)
    {
        if(resources[0][k].size() > 0){
            num_groups++;
            group_reqs_first.push_back(resources[0][k][0]);
            max_resource_id = std::max(max_resource_id, resources[0][k][0]->get_resource_id());
        }
    }
    sort_by_request_length(group_reqs_first);

    // Filter the second-longest (or null if none) into group_reqs_second
    std::vector<const RequestBound*> group_reqs_second(max_resource_id+1, NULL);
    for(k=0; k<resources[0].size(); k++)
        if (resources[0][k].size() > 1) {
            group_reqs_second[resources[0][k][0]->get_resource_id()] = resources[0][k][1];
        }

    // We need for each task the maximum request span.  We also need the
    // maximum direct blocking from remote partitions for each request. We
    // can determine both in one pass.

    unsigned int i;

    // Direct blocking results
    BlockingBounds* _results = new BlockingBounds(info);
    BlockingBounds& results = *_results;

    for (i = 0; i < info.get_tasks().size(); i++)
    {
        const TaskInfo& tsk  = info.get_tasks()[i];
        Interference bterm;

        if (tsk.get_requests().size() > 0)
        {
            const RequestBound* this_req = &tsk.get_requests()[0];

            Interference wblocking, wblocking_w1;

            int idx;
            unsigned int parallelism = procs_per_cluster;
            enumerate(resources, ct, idx)
            {
                parallelism = procs_per_cluster;

                if (idx == dedicated_irq)
                parallelism--;

                if (parallelism && (int) tsk.get_cluster() == idx)
                parallelism--;
            }
            // Calculate wblocking
            if (parallelism == 0)
            {
                wblocking = 0;
            }
            else
            {
                unsigned int count = 0;
                unsigned int total_length = 0;
                unsigned int num_remaining = std::min(num_groups,parallelism);

                // There is a special case where this request is the largest in its group,
                // but the next largest is small enough that some other group would be
                // included in the num_remaining selected - just in case, we'll track
                // the second largest in the group (if this one is the largest), and compare
                // at the end, taking either the (num_remaining+1)th largest or the second
                // largest in this group
                bool is_largest_in_group = false;
                unsigned int second_largest_length = 0;

                // Find num_remaining jobs that block me (group_reqs_first is ordered
                // by decreasing critical-section length)
                foreach(group_reqs_first, it)
                {
                    const RequestBound* req = *it;
                    if (num_remaining == 0 && is_largest_in_group == false) {
                        break;
                    }
                    else if (num_remaining == 0 && is_largest_in_group == true) {
                        // We've hit the special case!  Take the largest of the second-largest-in-this-group
                        // and the next largest group (corresponding to req) as the last
                        unsigned int num;
                        num = 1; // FIFO constraint means at most 1 request per task,
                                 // and num_remaining is >0 if we're here
                        if (num*req->get_request_length() > second_largest_length)
                        {
                            total_length -= second_largest_length;
                            total_length += num*req->get_request_length();
                        }
                        break;
                    }

                    if (req->get_task() != &tsk)
                    {
                        // This request is not me - go ahead and add it!
                        unsigned int num;
                        num = 1; // FIFO constraint means at most 1 request per task,
                                 // and num_remaining is >0 if we're here
                        total_length += num*req->get_request_length();
                        count += num;
                        num_remaining -= num;
                    }
                    else
                    {
                        // I am the longest request in my group!  Look up the
                        // second-longest and use that instead
                        const RequestBound *secondLongest = group_reqs_second[req->get_resource_id()];
                        if (secondLongest == NULL)
                            continue;

                        unsigned int num;
                        num = 1; // FIFO constraint means at most 1 request per task,
                                 // and num_remaining is >0 if we're here
                        second_largest_length = num*secondLongest->get_request_length();
                        total_length += second_largest_length;
                        is_largest_in_group = true;
                        count += num;
                        num_remaining -= num;
                    }
                }
                wblocking.total_length = total_length;
                wblocking.count = count;
            }

            wblocking_w1 = wblocking;
            wblocking_w1.total_length += this_req->get_request_length();
            wblocking_w1.count    += 1;

            // Combine
            results.raise_request_span(i, wblocking_w1);
            bterm += wblocking;
        }
        results[i] = bterm;
    } // foreach

    // This is the initial delay due to priority donation.
    charge_arrival_blocking(info, results);

    return _results;
}

